from __future__ import division, print_function
from base64 import encodebytes
from glob import glob
import requests
import pandas as pd
import json
import numpy as np
import multiprocessing
import re
import os
import io
import traceback
from time import sleep

from .utils import LightCurveCollection


class SkyPatrolClient:
    """
            The SkyPatrolClient allows users to interact with the ASAS-SN Sky Patrol photometry database.
        This client enables users to use ADQL, cone searches, random samples, and catalog ID lookups on the input catalogs.

        Queries to the input catalogs will either be returned as pandas DataFrames containing aggregate information
        on astronomical targets, or they will be returned as LightCurveCollections containing photometry data from all
        queried targets.
    """

    def __init__(self, verbose=True):
        """
        Creates SkyPatrolClient object
        """
        self.index = None
        self.verbose = verbose
        try:
            url = "http://asassn-lb01.ifa.hawaii.edu:9006/get_current_message"
            url_data = requests.get(url).content
            self.message = url_data.decode()
            self._verbose_print(self.message)

            url = "http://asassn-lb01.ifa.hawaii.edu:9006/get_schema"
            url_data = requests.get(url).content
            schema = json.loads(url_data)

            url = "http://asassn-lb01.ifa.hawaii.edu:9006/get_counts"
            url_data = requests.get(url).content
            counts = json.loads(url_data)

            url = "http://asassn-lb01.ifa.hawaii.edu:9006/get_block_servers"
            url_data = requests.get(url).content
            self.block_servers = json.loads(url_data)

            self.catalogs = SkyPatrolClient.InputCatalogs(schema, counts)

        except ConnectionError as e:
            raise ConnectionError("Unable to connect to ASAS-SN Servers")

    def adql_query(
        self, query_str, download=False, save_dir=None, file_format="parquet", threads=1
    ):
        """
        Query the ASAS-SN Sky Patrol Input Catalogs with an ADQL string.
        See README.md for more on accepted ADQL context and functions.

        :param query_str: ADQL query string
        :param download: return full light curves if True, return catalog information if False
        :param save_dir: if set, then light curves will write to file as they are downloaded
        :param file_format: format to save light curves ['parquet', 'pickle', 'csv']
        :param threads: number of real threads to use for pulling light curves from server.
        :return: if 'download' if False; pandas Dataframe containing catalog information of targets;
                else LightCurveCollection
        """
        # Check inputs
        if type(download) is not bool:
            raise ValueError("'download' must be boolean value")
        if type(threads) is not int:
            raise ValueError("'threads' must be integer value")
        if type(query_str) is not str:
            raise ValueError("'query_str' must me string value")

        # Trim ADQL input
        query_str = re.sub(r"(\s+)", " ", query_str)
        query_hash = encodebytes(bytes(query_str, encoding="utf-8")).decode()

        # Query Flask API with SQL bytes
        url = f"http://asassn-lb01.ifa.hawaii.edu:9006/lookup_sql/{query_hash}"
        response = requests.post(url, json={"format": "arrow", "download": download})

        self._validate_response(response)

        # Deserialize from arrow
        tar_df = _deserialize(response.content)
        self.index = tar_df

        if download is False:
            # Return a dataframe of catalog search results
            return tar_df
        else:
            if save_dir:
                self._save_index(save_dir, file_format)
            # Get lightcurve ids to pull
            if "asas_sn_id" in tar_df.columns:
                id_col = "asas_sn_id"
                catalog = "extrasolar"
            elif "comets" in query_str:
                id_col = "name"
                catalog = "comets"
            elif "asteroids" in query_str:
                id_col = "name"
                catalog = "asteroids"
            else:
                raise ValueError("needs propper id column to download lightcurves")
            # Get tar ids
            tar_ids = list(id_col)

            # Returns a LightCurveCollection object, or a list of light curve files when save_dir is set
            return self._get_curves(
                query_hash, tar_ids, catalog, save_dir, file_format, threads
            )

    def cone_search(
        self,
        ra_deg,
        dec_deg,
        radius,
        units="deg",
        catalog="master_list",
        cols=None,
        download=False,
        save_dir=None,
        file_format="parquet",
        threads=1,
    ):
        """
        Query the ASAS-SN Sky Patrol Input Catalogs for all targets within a cone of the sky.
        Does NOT return solar system targets (i.e. asteroids and coma).

        :param ra_deg: right ascension of cone.
                       accepts degree-decimal (float) or HH:MM:SS(.SS) sexagesimal (string)
        :param dec_deg: declination of cone.
                        accepts degree/decimal as float or DD:MM:SS(.SS) sexagesimal (string)
        :param radius: radius in degrees of cone, float
        :param units: units for cone radius.
                     'deg': degree decimal
                     'arcmin': arcminutes
                     'arcsec' arcseconds
        :param catalog: which catalog are we searching
        :param cols: columns to return from the given input catalog;
                     if None default = ['asas_sn_id', 'ra_deg', 'dec_deg']
        :param download: return full light curves if True, return catalog information if False
        :param save_dir: if set, then light curves will write to file as they are downloaded
        :param file_format: format to save light curves ['parquet', 'pickle', 'csv']
        :param threads: number of real threads to use for pulling light curves from server.
        :return: if 'download' if False; pandas Dataframe containing catalog information of targets;
                else LightCurveCollection
        """
        # Check inputs
        if type(download) is not bool:
            raise ValueError("'download' must be boolean value")
        if type(threads) is not int:
            raise ValueError("'threads' must be integer value")
        if catalog not in self.catalogs.catalog_names() or catalog in [
            "asteroids",
            "comets",
        ]:
            raise ValueError("invalid catalog")
        if cols is None:
            cols = ["asas_sn_id", "ra_deg", "dec_deg"]
            if catalog == "m_giants":
                cols.append("gaia_id")
            elif catalog == "master_list":
                cols.append("catalog_sources")
            elif catalog not in ["stellar_main", "master_list"]:
                cols.append("name")
        if set(cols).issubset(set(self.catalogs[catalog]["col_names"])) is False:
            raise ValueError(
                "one or more of the listed cols is not a column name in the selected catalog"
            )

        # Change units
        if units != "deg":
            radius = _arc_to_deg(np.float(radius), units)

        # Query the Flask server API for cone
        url = f"http://asassn-lb01.ifa.hawaii.edu:9006/lookup_cone/radius{radius}_ra{ra_deg}_dec{dec_deg}"
        response = requests.post(
            url,
            json={
                "catalog": catalog,
                "cols": cols,
                "format": "arrow",
                "download": download,
            },
        )

        # Check response
        self._validate_response(response)

        # Deserialize from arrow
        tar_df = _deserialize(response.content)
        self.index = tar_df

        if download is False:
            # Return a dataframe of catalog search results
            return tar_df
        else:
            if save_dir:
                self._save_index(save_dir, file_format)
            # Generate query information
            query_id = (
                f"conectr-{radius}_conera-{ra_deg}_conedec-{dec_deg}|catalog-{catalog}|cols-"
                + "/".join(cols)
            )
            query_hash = encodebytes(bytes(query_id, encoding="utf-8")).decode()

            # Get lightcurve ids to pull
            id_col = "asas_sn_id" if catalog not in ["asteroids", "comets"] else "name"
            tar_ids = list(tar_df[id_col])

            # Returns a LightCurveCollection object, or a list of light curve files when save_dir is set
            return self._get_curves(
                query_hash, tar_ids, catalog, save_dir, file_format, threads
            )

    def query_list(
        self,
        target_ids,
        id_col="asas_sn_id",
        catalog="master_list",
        cols=None,
        download=False,
        save_dir=None,
        file_format="parquet",
        threads=1,
    ):
        """
        Query the ASAS-SN Sky Patrol Input Catalogs for all targets with the given identifiers.
        to view the available list of catalogs and identifiers see SkyPatrolClient.catalogs.

        Most of our astronomical targets are in the stellar_main catalog, which has been pre-crossmatched to GaiaDR2,
        ATLAS Refcat2, SDSS, AllWISE, and Tess Input Catalog (TIC v8).

        Thus searching for light curves with a list of Gaia IDs would require catalog='stellar_main', id_col='gaia_id'.
        Our other input catalogs should be searched with id_col='name', or by columns giving alternative ids.

        :param target_ids: list of target ids for query; list
        :param id_col: the column on the given catalog to search against; string
        :param catalog: which catalog are we searching
        :param cols: columns to return from the given input catalog;
                     if None default = ['asas_sn_id', 'ra_deg', 'dec_deg']
        :param download: return full light curves if True, return catalog information if False
        :param save_dir: if set, then light curves will write to file as they are downloaded
        :param file_format: format to save light curves ['parquet', 'pickle', 'csv']
        :param threads: number of real threads to use for pulling light curves from server.
        :return: if 'download' if False; pandas Dataframe containing catalog information of targets;
                else LightCurveCollection
        """
        # Check inputs
        if type(download) is not bool:
            raise ValueError("'download' must be boolean value")
        if type(threads) is not int:
            raise ValueError("'threads' must be integer value")
        # Limit sources
        if catalog not in self.catalogs.catalog_names():
            raise ValueError("invalid catalog")

        # Check catalog columns
        if id_col not in list(self.catalogs[catalog]["col_names"]):
            raise ValueError("invalid column")
        # Set default columns
        if cols is None and catalog not in ["asteroids", "comets"]:
            cols = ["asas_sn_id", "ra_deg", "dec_deg"]
            if catalog == "m_giants":
                cols.append("gaia_id")
            elif catalog == "master_list":
                cols.append("catalog_sources")
            elif catalog not in ["stellar_main", "master_list"]:
                cols.append("name")
        if cols is None and catalog in ["asteroids", "comets"]:
            cols = ["name"]
        # Ensure valid columns
        if set(cols).issubset(set(self.catalogs[catalog]["col_names"])) is False:
            raise ValueError(
                "one or more of the listed cols is not a column name in the selected catalog"
            )
        if id_col not in cols:
            cols.append(id_col)

        if type(target_ids) in [str, int]:
            target_ids = [target_ids]

        # Query API with list (POST METHOD)
        url = f"http://asassn-lb01.ifa.hawaii.edu:9006/lookup_targets/catalog_list"
        response = requests.post(
            url,
            json={
                "tar_ids": target_ids,
                "catalog": catalog,
                "id_col": id_col,
                "cols": cols,
                "format": "arrow",
                "download": download,
            },
        )

        # Check response
        self._validate_response(response)

        # Deserialize from arrow
        tar_df = _deserialize(response.content)
        self.index = tar_df

        if download is False:
            # Return a dataframe of catalog search results
            return tar_df
        else:
            if save_dir:
                self._save_index(save_dir, file_format)

            # Generate query information
            query_id = (
                f"listlen-{len(target_ids)}_listfirst-{target_ids[0]}_listend-{target_ids[-1]}"
                f"|catalog-{catalog}|id_col-{id_col}|cols-" + "/".join(cols)
            )
            query_hash = encodebytes(bytes(query_id, encoding="utf-8")).decode()

            # Get lightcurve ids to pull
            id_col = "asas_sn_id" if catalog not in ["asteroids", "comets"] else "name"
            tar_ids = list(tar_df[id_col])

            # Returns a LightCurveCollection object, or a list of light curve files when save_dir is set
            return self._get_curves(
                query_hash, tar_ids, catalog, save_dir, file_format, threads
            )

    def random_sample(
        self,
        n,
        catalog="master_list",
        cols=None,
        download=False,
        save_dir=None,
        file_format="parquet",
        threads=1,
    ):
        """
        Get n random targets from the ASAS-SN Sky Patrol Input Catalogs.

        :param n: number of targets to randomly sample
        :param catalog: which catalog are we searching
        :param cols: columns to return from the given input catalog;
                     if None default = ['asas_sn_id', 'ra_deg', 'dec_deg']
        :param download: return full light curves if True, return catalog information if False
        :param save_dir: if set, then light curves will write to file as they are downloaded
        :param file_format: format to save light curves ['parquet', 'pickle', 'csv']
        :param threads: number of real threads to use for pulling light curves from server.
        :return: if 'download' if False; pandas Dataframe containing catalog information of targets;
                else LightCurveCollection
        """
        # Check inputs
        if type(download) is not bool:
            raise ValueError("'download' must be boolean value")
        if type(threads) is not int:
            raise ValueError("'threads' must be integer value")
        # Valid catalogs
        if catalog not in self.catalogs.catalog_names():
            raise ValueError("invalid catalog")
        # Set default columns
        if cols is None and catalog not in ["asteroids", "comets"]:
            cols = ["asas_sn_id", "ra_deg", "dec_deg"]
            if catalog == "m_giants":
                cols.append("gaia_id")
            elif catalog == "master_list":
                cols.append("catalog_sources")
            elif catalog not in ["stellar_main", "master_list"]:
                cols.append("name")
        if cols is None and catalog in ["asteroids", "comets"]:
            cols = ["name"]
        # Ensure valid columns
        if set(cols).issubset(set(self.catalogs[catalog]["col_names"])) is False:
            raise ValueError(
                "one or more of the listed cols is not a column name in the selected catalog"
            )

        # Limit sample length
        if n >= 1000000:
            raise ValueError("max sample size is 1 million")

        # Query API with list (POST METHOD)
        url = f"http://asassn-lb01.ifa.hawaii.edu:9006/lookup_targets/random_{n}"
        response = requests.post(
            url,
            json={
                "catalog": catalog,
                "cols": cols,
                "format": "arrow",
                "download": download,
            },
        )

        # Check response
        self._validate_response(response)

        # Deserialize from arrow
        tar_df = _deserialize(response.content)
        self.index = tar_df

        if download is False:
            # Return a dataframe of catalog search results
            return tar_df
        else:
            if save_dir:
                self._save_index(save_dir, file_format)
            # Get lightcurve ids to pull
            id_col = "asas_sn_id" if catalog not in ["asteroids", "comets"] else "name"
            tar_ids = list(tar_df[id_col])

            # Generate query information
            query_id = f"random-{n}|catalog-{catalog}|cols-" + "/".join(cols)
            query_hash = encodebytes(bytes(query_id, encoding="utf-8")).decode()

            # Returns a LightCurveCollection object, or a list of light curve files when save_dir is set
            return self._get_curves(
                query_hash, tar_ids, catalog, save_dir, file_format, threads
            )

    def _get_curves(
        self, query_hash, tar_ids, catalog, save_dir, file_format, threads=1
    ):
        # Get number of id chunks
        n_chunks = int(np.ceil(len(tar_ids) / 1000))
        self._verbose_print("Downloading Curves...")
        # Get targets via mutlithreaded requests
        with multiprocessing.Pool(processes=threads) as pool:
            results = [
                pool.apply_async(
                    self._get_lightcurve_chunk,
                    args=(query_hash, idx, catalog, save_dir, file_format,),
                )
                for idx in range(n_chunks)
            ]

            chunks = []
            count = 0
            for r in results:
                data, n = r.get()
                chunks.append(data)
                count += n
                self._verbose_print(
                    f"Pulled {count:,} of {len(self.index):,}", end="\r"
                )

            self._verbose_print("")
        if save_dir is not None:
            # Return list of filenames
            return [file for chunk in chunks for file in chunk]
        else:
            # Return in memory data as LightCurveCollection
            data = pd.concat(chunks)
            id_col = "asas_sn_id" if catalog not in ["asteroids", "comets"] else "name"
            return LightCurveCollection(data, self.index, id_col)

    def _get_lightcurve_chunk(
        self, query_hash, block_idx, catalog, save_dir, file_format
    ):
        # Available servers
        n_servers = len(self.block_servers)
        # Start timeout at 0
        timeout = 1
        # Download server (for balance)
        server_idx = block_idx % n_servers
        # Success flag
        success = False

        # Loop through until we download or timeout
        while success is False and timeout <= 5:
            try:            
                # Query API with (POST METHOD)
                url = (
                    f"http://{self.block_servers[server_idx]}:9006/get_block/"
                    f"query_hash-{query_hash}-block_idx-{block_idx}-catalog-{catalog}"
                )
                response = requests.get(url)

                # Pandas dataframe
                data = _deserialize(response.content)
                # ID and count
                id_col = "asas_sn_id" if catalog not in ["asteroids", "comets"] else "name"
                count = len(data[id_col].unique())

                success = True
            except:
                sleep(timeout)
                server_idx = (server_idx + 1) % n_servers
                
                # Raise timeout if we've tried all servers once
                if (server_idx % n_servers) == (block_idx % n_servers):
                    timeout += 1
        
        # If download fails, raise exception 
        if success is False:
            raise TimeoutError("Lightcurve servers unavailable, try again later")

        # Write to disk or return in memory
        if save_dir is not None:
            lcs = LightCurveCollection(data, self.index, id_col)
            return lcs.save(save_dir, file_format, include_index=False), count
        else:
            return data, count

    def _save_index(self, save_dir, file_format):
        if file_format == "parquet":
            self.index.to_parquet(os.path.join(save_dir, "index.parq"))
        elif file_format == "pickle":
            self.index.to_pickle(os.path.join(save_dir, "index.pkl"))
        elif file_format == "csv":
            self.index.to_csv(os.path.join(save_dir, "index.csv"), index=False)
        else:
            raise ValueError(
                f"invalid format: '{file_format}' not in ['parquet', 'csv', 'pickle']"
            )

    def _verbose_print(self, msg, end="\n"):
        if self.verbose:
            print(msg, end=end, flush=True)

    def _validate_response(self, response):
        """
        Validate the response before deserialization
        """
        # If the response is not an error (4xx, 5xx), it is true
        if not response:
            response.raise_for_status()

    class InputCatalogs(object):
        """
        Data structure for holding metadata on ASAS-SN Sky Patrol's input catalogs.

        Stores input catalog names as well as their searchable columns and total number of targets.
        """

        def __init__(self, schema, counts):
            for name, col_data in schema.items():
                self.__dict__[name] = pd.DataFrame(col_data)
            self.counts = counts

        def __str__(self):
            rep_str = "\n"
            for table_name, df in self.__dict__.items():
                if table_name == "counts":
                    continue
                rep_str += (
                    f"Table Name:  {table_name}\n"
                    f"Num Columns: {len(df)}\n"
                    f"Num Targets: {self.counts[table_name]}"
                    f"{df.head(10)}\n\n"
                )
            return rep_str

        def __repr__(self):
            rep_str = "\n"
            for table_name, df in self.__dict__.items():
                if table_name == "counts":
                    continue
                rep_str += (
                    f"Table Name:  {table_name}\n"
                    f"Num Columns: {len(df)}\n"
                    f"Num Targets: {self.counts[table_name]}\n\n"
                )
            return rep_str

        def catalog_names(self):
            """
            Get all the names of our available input catalogs

            :return: names of all input catalogs (list)
            """
            return self.__dict__.keys()

        def __getitem__(self, item):
            return self.__dict__[item]


def _deserialize(buffer):
    # Deserialize from bytes
    return pd.read_parquet(io.BytesIO(buffer))


def _arc_to_deg(arc, unit):
    """
    Convert arc minutes or seconds to decimal degree units

    :param arc: length of arc, float or np array
    :param unit: 'arcmin' or 'arcsec'
    :return: decimal degrees
    """
    if unit == "arcmin":
        return arc / 60
    elif unit == "arcsec":
        return arc / 3600
    else:
        raise ValueError("unit not in ['arcmin', 'arcsec']")


def load_collection(save_dir, file_format="parquet"):
    """
    Loads a LightCurveCollection from directory.
    Requires an index and light curve files saved from previous collection
    :param save_dir: path where collection is saved
    :param file_format: format of saved light curves
    :return: LightCurveCollection
    """
    if file_format == "parquet":
        index = pd.read_parquet(os.path.join(save_dir, "index.parq"))
        files = glob(os.path.join(save_dir, "*.parq"))
        lc_files = list(filter(lambda i: os.path.basename(i) != "index.parq", files))
        data = pd.concat(pd.read_parquet(file) for file in lc_files)

    elif file_format == "pickle":
        index = pd.read_pickle(os.path.join(save_dir, "index.pkl"))
        files = glob(os.path.join(save_dir, "*.pkl"))
        lc_files = list(filter(lambda i: os.path.basename(i) != "index.pkl", files))
        data = pd.concat(pd.read_pickle(file) for file in lc_files)

    elif file_format == "csv":
        index = pd.read_csv(os.path.join(save_dir, "index.csv"))
        files = glob(os.path.join(save_dir, "*.csv"))
        lc_files = list(filter(lambda i: os.path.basename(i) != "index.csv", files))
        data = pd.concat(pd.read_csv(file) for file in lc_files)

    else:
        raise ValueError(
            f"invalid format: '{file_format}' not in ['parquet', 'csv', 'pickle']"
        )

    id_col = "asas_sn_id" if "asas_sn_id" in index.columns else "name"
    return LightCurveCollection(data, index, id_col)



from __future__ import division, print_function
import os
from astropy.timeseries import LombScargle
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from .wavelet import LS_wavelet


class LightCurveCollection(object):
    """
    Object for storing and analysing ASAS-SN Sky Patrol light curves.
    Returned by any SkyPatrolClient query where download=True
    """

    def __init__(self, data, catalog_info, id_col):

        self.id_col = id_col
        self.data = data
        self.catalog_info = catalog_info
        # Add filter column
        Vcams = ['ba', 'bb', 'bc', 'bd', 'be', 'bf', 'bg', 'bh']
        gcams = ['bA', 'bB', 'cB', 'bC', 'bD', 'bE', 'bF', 'bG', 'bH',
                 'bi', 'bj', 'bk', 'bl', 'bm', 'bn', 'bo', 'bp', 'bq', 'br', 'bs', 'bt']
        self.data['phot_filter'] = self.data.camera.apply(lambda x: 'V' if x in Vcams else 'g' if x in gcams else None)
        self.data = self.data[self.data['phot_filter'].notna()]

    def apply_function(
        self, func, col="mag", include_non_det=False, include_poor_images=False, phot_filter='all'
    ):
        """
        Apply a custom aggregate function to all light curves in the collection.

        :param func: custom aggregate function
        :param col: column to apply aggregate function; defaluts to 'mag'
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :param include_poor_images whether or not to include images of poor or unknown quality; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to all
        :return: pandas Dataframe with results
        """

        # Filter preferences for this function call only
        data = self.data
        if not include_non_det:
            data = data[data["mag_err"] < 99]
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Filter by filter
        if phot_filter == 'g':
            data = data[data['phot_filter'] == 'g']
        elif phot_filter == 'V':
            data = data[data['phot_filter'] == 'V']
        elif phot_filter == 'all':
            pass
        else:
            raise ValueError("phot_filter must be in ['g', 'V', 'all']")

        return data.groupby(self.id_col).agg({col: func})

    def stats(
            self, include_non_det=False, include_poor_images=False, phot_filter='all'
    ):
        """
        Calculate simple aggregate statistics on the collection.
        Gets the mean and stddev magnitude for each curve as well as the total number of epochs observed.

        :param include_poor_images: whether or not to include images of poor or unknown quality; defaults to False
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to all
        :return: pandas Dataframe with results
        """
        # Filter preferences for this function call only
        data = self.data
        if not include_non_det:
            data = data[data["mag_err"] < 99]
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Filter by filter
        if phot_filter == 'g':
            data = data[data['phot_filter'] == 'g']
        elif phot_filter == 'V':
            data = data[data['phot_filter'] == 'V']
        elif phot_filter == 'all':
            pass
        else:
            raise ValueError("phot_filter must be in ['g', 'V', 'all']")

        return data.groupby(self.id_col).agg(
            mean_mag=("mag", "mean"), std_mag=("mag", "std"), epochs=("mag", "count")
        )

    def __getitem__(self, item):
        if type(item) == pd.Series or type(item) == list or type(item) == np.ndarray:
            data = self.data[self.data[self.id_col] in item]
            catalog_info = self.catalog_info[self.catalog_info[self.id_col] in item]
            return LightCurveCollection(data, catalog_info, self.id_col)

        else:
            return self.__get_lc(item)

    def __get_lc(self, key):
        source = self.catalog_info[self.id_col] == key
        meta = self.catalog_info[source]
        data = self.data[self.data[self.id_col] == key]
        return LightCurve(data, meta)

    def itercurves(self):
        """
        Generator to iterate through all light curves in the collection.
        :return: a generator that iterates over the collection
        """
        for key, data in self.data.groupby(self.id_col):
            source = self.catalog_info[self.id_col] == key
            meta = self.catalog_info[source]
            yield LightCurve(data, meta)

    def __len__(self):
        return len(self.catalog_info)

    def save(
            self, save_dir, file_format="parquet", include_index=True
    ):
        """
        Saves entire light curve collection to a given directory.

        :param save_dir: directory name
        :param file_format: file format of saved objects ['parquet', 'csv', 'pickle']
        :param include_index: whether or not to save index (catalog_info)
        :return: a list of file names
        """
        filenames = []
        if file_format == "parquet":
            if include_index:
                self.catalog_info.to_parquet(os.path.join(save_dir, "index.parq"))
                filenames.append("index.parq")
            for lc in self.itercurves():
                file = os.path.join(save_dir, f"{lc.meta[self.id_col].values[0]}.parq")
                lc.save(file, file_format="parquet")
                filenames.append(file)

        elif file_format == "pickle":
            if include_index:
                self.catalog_info.to_pickle(os.path.join(save_dir, "index.pkl"))
                filenames.append("index.pkl")
            for lc in self.itercurves():
                file = os.path.join(save_dir, f"{lc.meta[self.id_col].values[0]}.pkl")
                lc.save(file, file_format="pickle")
                filenames.append(file)

        elif file_format == "csv":
            if include_index:
                self.catalog_info.to_csv(
                    os.path.join(save_dir, "index.csv"), index=False
                )
                filenames.append("index.csv")
            for lc in self.itercurves():
                file = os.path.join(save_dir, f"{lc.meta[self.id_col].values[0]}.csv")
                lc.save(file, file_format="csv")
                filenames.append(file)
        else:
            raise ValueError(
                f"invalid format: '{file_format}' not in ['parquet', 'csv', 'pickle']"
            )

        return filenames


class LightCurve:
    """
    Object for analysing and visualizing ASAS-SN Sky Patrol light curves.
    """

    def __init__(self, data, meta):

        self._validate(data)
        self.data = data
        self.meta = meta
        self.epochs = len(data)

    @staticmethod
    def _validate(obj):
        # verify there is a column latitude and a column longitude
        if (
            "jd" not in obj.columns
            or "mag" not in obj.columns
            or "mag_err" not in obj.columns
        ):
            raise AttributeError("Must have 'jd'(julian date), 'mag' and 'mag_err'")

    def save(
            self, filename, file_format="parquet"
    ):
        """
        Save the light curve to csv.
        :param filename: filename to save this light curve.
        :param file_format: file format of saved objects ['parquet', 'csv', 'pickle']
        :return: void
        """
        if file_format == "parquet":
            self.data.to_parquet(filename)
        elif file_format == "pickle":
            self.data.to_pickle(filename)
        elif file_format == "csv":
            with open(filename, "w+") as f:
                f.write(f"# {self.meta.to_json(orient='records')[2:-2]}\n")
            self.data.to_csv(filename, mode="a", index=False)
        else:
            raise ValueError(
                f"invalid format: '{file_format}' not in ['parquet', 'csv', 'pickle']"
            )

    def plot(
            self,
            figsize=(12, 8),
            font_size=10,
            save_file=None,
            include_poor_images=False,
            include_non_det=True,
            phot_filter='all',
    ):
        """
        Plots the given light curve with error bars.

        :param figsize: size of the plot.
        :param font_size: font size for the plotting.
        :param save_file: file name to save the plot; if None plot will be directly displayed
        :param include_poor_images: whether or not to include images of poor or unknown quality; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to all
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :return: void
        """
        # Filter preferences
        data = self.data

        # Filter out pool quality images
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Filter detections
        errors = data.mag_err > 99
        detections = data[~errors]

        # Plot
        plt.figure(figsize=figsize)

        # Set font size
        plt.rcParams.update({'font.size': font_size})
        # Diff colors for filters
        if phot_filter in ['g', 'all']:
            plt.errorbar(
                x=detections[detections['phot_filter'] == 'g'].jd - 2450000,
                y=detections[detections['phot_filter'] == 'g'].mag,
                yerr=detections[detections['phot_filter'] == 'g'].mag_err,
                fmt="o",
                c="mediumblue",
                label="g band",
            )
        if phot_filter in ['V', 'all']:
            plt.errorbar(
                x=detections[detections['phot_filter'] == 'V'].jd - 2450000,
                y=detections[detections['phot_filter'] == 'V'].mag,
                yerr=detections[detections['phot_filter'] == 'V'].mag_err,
                fmt="o",
                c="teal",
                label="V band",
            )
        if phot_filter not in ['g', 'V', 'all']:
            raise ValueError("phot_filter must be in ['g', 'V', 'all']")

        # Plot non-detections
        if include_non_det:
            plt.errorbar(
                x=data[errors].jd - 2450000,
                y=data[errors].mag,
                fmt="v",
                c="red",
                label="non-detections",
            )
        # Label plots
        self._label_plots(font_size)
        plt.legend()
        plt.xlabel("Date (JD-2450000)")
        plt.ylabel("Magnitude")
        plt.gca().invert_yaxis()

        if save_file:
            plt.savefig(save_file)
        else:
            plt.show()

        # Reset font size (default = 10)
        plt.rcParams.update({'font.size': 10})

    def lomb_scargle(
        self,
        fit_mean=True,
        center_data=True,
        nterms=1,
        normalization="standard",
        minimum_frequency=0.001,
        maximum_frequency=25,
        method="auto",
        samples_per_peak=5,
        nyquist_factor=5,
        plot=True,
        figsize=(12, 8),
        font_size=10,
        save_file=None,
        include_poor_images=False,
        include_non_det=False,
        phot_filter='all'
    ):
        """
        Thin wrapper around the astropy LombScargle utility to determine frequency and power spectra of the given
        light curve. Default values work for most variable sources.

        :param fit_mean: if True, include a constant offset as part of the model at each frequency. This can lead to more accurate results, especially in the case of incomplete phase coverage.
        :param center_data: if True, pre-center the data by subtracting the weighted mean of the input data. This is especially important if fit_mean = False
        :param nterms: number of terms to use in the Fourier fit
        :param normalization: Normalization to use for the periodogram.
        :param minimum_frequency: If specified, then use this minimum frequency rather than one chosen based on the size of the baseline.
        :param maximum_frequency: If specified, then use this maximum frequency rather than one chosen based on the average nyquist frequency.
        :param method: specify the lomb scargle implementation to use. Options are:

            -  ‘auto’: choose the best method based on the input

            -  ‘fast’: use the O[N log N] fast method. Note that this requires evenly-spaced frequencies: by default this will be checked unless assume_regular_frequency is set to True.

            -  ‘slow’: use the O[N^2] pure-python implementation

            -  ‘cython’: use the O[N^2] cython implementation. This is slightly faster than method=’slow’, but much more memory efficient.

            -  ‘chi2’: use the O[N^2] chi2/linear-fitting implementation

            -  ‘fastchi2’: use the O[N log N] chi2 implementation. Note that this requires evenly-spaced frequencies: by default this will be checked unless assume_regular_frequency is set to True.

            -  ‘scipy’: use scipy.signal.lombscargle, which is an O[N^2] implementation written in C. Note that this does not support heteroskedastic errors.

        :param samples_per_peak: The approximate number of desired samples across the typical peak
        :param nyquist_factor: The multiple of the average nyquist frequency used to choose the maximum frequency if maximum_frequency is not provided.
        :param plot: if True, then the function also produces a plot of the power spectrum.
        :param figsize: size of the plot.
        :param font_size: font size for the plotting.
        :param save_file: file name to save the plot; if None plot will be directly displayed
        :param include_poor_images: whether or not to include images of poor or unknown quality; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to all
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :return: power, frequency and the astropy LombScargle object
        """
        # Filter preferences for this function call only
        data = self.data
        if not include_non_det:
            data = data[data["mag_err"] < 99]
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Filter by filter
        if phot_filter == 'g':
            data = data[data['phot_filter'] == 'g']
        elif phot_filter == 'V':
            data = data[data['phot_filter'] == 'V']
        elif phot_filter == 'all':
            pass
        else:
            raise ValueError("phot_filter must be in ['g', 'V', 'all']")

        ls = LombScargle(
            data.jd,
            data.mag,
            fit_mean=fit_mean,
            center_data=center_data,
            nterms=nterms,
            normalization=normalization,
        )

        frequency, power = ls.autopower(
            minimum_frequency=minimum_frequency,
            maximum_frequency=maximum_frequency,
            method=method,
            samples_per_peak=samples_per_peak,
            nyquist_factor=nyquist_factor,
            normalization=normalization,
        )
        if plot:
            plt.figure(figsize=figsize)
            # Set font size
            plt.rcParams.update({'font.size': font_size})

            plt.plot(frequency, power)
            self._label_plots(font_size)
            plt.xlabel("Frequency")
            plt.ylabel("Power")

            if save_file:
                plt.savefig(save_file)
            else:
                plt.show()

            # Reset font size (default = 10)
            plt.rcParams.update({'font.size': 10})

        return frequency, power, ls

    def wavelet_power(
        self,
        tt,
        ff,
        include_poor_images=False,
        include_non_det=True,
        phot_filter='all',
        tradeoff=2
        ):
        '''
        Constructs a wavelet-transform power spectrum.

        :param tt: Array of times at which to evaluate wavelet PS
        :param ff: Array of frequencies at which to evaluate wavelet PS
        :param include_poor_images: whether or not to include images of poor or unknown quality; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to g
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :tradeoff: Tradeoff parameter between frequency and time resolution

        :return: numpy array containing wavelet power at provided times and frequencies.
        '''

        # Filter preferences for this function call only
        data = self.data

        if not include_non_det:
            data = data[data["mag_err"] < 99]
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Filter by filter
        if phot_filter == 'g':
            data = data[data['phot_filter'] == 'g']
        elif phot_filter == 'V':
            data = data[data['phot_filter'] == 'V']
        elif phot_filter == 'all':
            pass
        else:
            raise ValueError("phot_filter must be in ['g', 'V', 'all']")

        # Data for wavelet
        x = data.jd
        y = data.mag
        e_y = data.mag_err

        return LS_wavelet(tt, ff, x, y, e_y, Γ=tradeoff)

    def find_period(
        self,
        frequency,
        power,
        best_frequency=None,
        plot=True,
        reference_epoch='max',
        figsize=(12, 8),
        font_size=10,
        save_file=None,
        include_poor_images=False,
        include_non_det=False,
        phot_filter='all'
    ):
        """
        Find the period of the light curve given the power spectrum produced by lomb_scargle.
        Also produces a phase-folded plot of the light curve.

        :param frequency: frequency from lomb_scargle()
        :param power: power from lomb_scargle()
        :param best_frequency: peak frequency for phase folding the light curve
        :param plot: if True, then the function also produces a plot of the phase-folded light curve
        :param reference_epoch: either 'max' or 'min', corresponding to the time of maximum/minimum flux in the light curve
        :param figsize: size of the plot
        :param font_size: font size for the plotting.
        :param save_file: file name to save the plot; if None plot will be directly displayed
        :param include_poor_images: whether or not to include images of poor or unknown quality; defaults to False
        :param include_non_det: whether or not to include non-detection events in analysis; defaults to False
        :param phot_filter: specify bandpass filter for photometry, either g, V, or all, defaults to all
        :return: period of the light curve
        """
        # Filter preferences for this function call only
        data = self.data
        if not include_non_det:
            data = data[data["mag_err"] < 99]
        if not include_poor_images:
            data = data[data["quality"] == "G"]

        # Get frequency
        if best_frequency is None:
            best_frequency = frequency[np.argmax(power)]

        # Get inverse for period and fold on time-space
        period = 1 / best_frequency

        if plot:
            plt.figure(figsize=figsize)
            # Set font size
            plt.rcParams.update({'font.size': font_size})

            if phot_filter in ['g', 'all']:
                # Filter for filter
                plot_data = data[data['phot_filter'] == 'g']

                # Get reference epoch for phasing
                if reference_epoch=='max':
                    ref_epoch=plot_data['jd'][plot_data.mag.idxmin()]
                elif reference_epoch=='min':
                    ref_epoch=plot_data['jd'][plot_data.mag.idxmax()]
                else:
                    ref_epoch=0.0

                phase=((plot_data.jd-ref_epoch)/period)%1
                # Concatenate for multiple peaks
                x = np.concatenate([phase, phase + 1])
                y = np.concatenate([plot_data.mag, plot_data.mag])
                plt.scatter(x, y, c='mediumblue', label='g band')

            if phot_filter in ['V', 'all']:
                # Filter for filter
                plot_data = data[data['phot_filter'] == 'V']

                # Get reference epoch for phasing
                if reference_epoch=='max':
                    ref_epoch=plot_data['jd'][plot_data.mag.idxmax()]
                elif reference_epoch=='min':
                    ref_epoch=plot_data['jd'][plot_data.mag.idxmin()]
                else:
                    ref_epoch=0.0

                phase=((plot_data.jd-ref_epoch)/period)%1
                # Concatenate for multiple peaks
                x = np.concatenate([phase, phase + 1])
                y = np.concatenate([plot_data.mag, plot_data.mag])
                plt.scatter(x, y, c='teal', label='V band')
            if phot_filter not in ['g', 'V', 'all']:
                raise ValueError("phot_filter must be in ['g', 'V', 'all']")

            # Create labels
            self._label_plots(font_size)
            plt.xlabel("Phase")
            plt.ylabel("Magnitude")
            plt.gca().invert_yaxis()

            if save_file:
                plt.savefig(save_file)
            else:
                plt.show()
        # Reset font size (default = 10)
        plt.rcParams.update({'font.size': 10})

        return period

    def _label_plots(self, font_size):
        suptitle = ""
        title = ""
        if "asas_sn_id" in self.meta.columns:
            suptitle += f"SkyPatrol ID: {self.meta.asas_sn_id.item()}"
        if "name" in self.meta.columns:
            suptitle += f"\nSource Name: {self.meta.name.item()}"
        if "ra_deg" in self.meta.columns:
            title += f"Right Ascention: {self.meta.ra_deg.item():.05f}"
        if "dec_deg" in self.meta.columns:
            title += f"\nDeclination: {self.meta.dec_deg.item():.05f}"

        title += f"\nEpochs: {self.epochs}"
        plt.title(title, loc="left", fontsize=font_size - 2)
        plt.suptitle(suptitle, fontsize=font_size + 2)
        plt.grid()


def merge(lcs, name):
    """
    Merge a LightCurveCollection or list of LightCurves to a single object.
    Useful for solar system objects with multiple designations.
    :param lcs: LightCurveCollection or list of LightCurves
    :param name: new name of the object
    :return: LightCurve
    """
    # Get all phot data
    if type(lcs) is list:
        lcs_data = [lc.data for lc in lcs]
    else:
        lcs_data = [lc.data for lc in lcs.itercurves()]

    return LightCurve(
        data=pd.concat(lcs_data),
        meta=pd.DataFrame({'name': [name]})
    )



#!/usr/bin/env python

import os
import sys
from setuptools import setup, Extension

# Prepare and send a new release to PyPI
# if "release" in sys.argv[-1]:
#     os.system("python setup.py sdist bdist_wheel")
#     os.system("twine upload dist/*")
#     os.system("rm -rf dist/skypatrol*")
#     sys.exit()

# Load the __version__ variable without importing the package already
# exec(open("skypatrol/version.py").read())

# Get dependencies
with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name="skypatrol",
    version="0.6.12",
    url="https://github.com/asas_sn/skypatrol/",
    author="ASAS-SN",
    author_email="kylehart@hawaii.edu",
    license="GPL v.3",
    packages=["pyasassn"],
    install_requires=install_requires,
    zip_safe=False,
)



import numpy as np
from astropy.timeseries import LombScargle
from tqdm.auto import tqdm

def LS_wavelet(tt, ff, x, y, e_y, Γ=2):

    """
    Computes a wavelet power spectrum. This is a *SLOW* implementation,
    hopefully to be replaced by something more efficient eventually.

    The units of tt and ff are assumed to be such that t * f is dimensionless.
    tt and x are assumed to have the same units.

    :param tt: Array of times at which to evaluate wavelet PS
    :param ff: Array of frequencies at which to evaluate wavelet PS
    :param x: Time axis of input time series
    :param y: Dynamical quantity (e.g. fluxes) of input time series
    :param e_y: Measurement errors of input time series
    :param Γ: tradeoff parameter between frequency and time resolution
              (by Fourier uncertainty principle). Larger values give
              better frequency resolution.

    :return: A numpy array containing the wavelet power spectrum.
    """

    # Instantiate accumulating array

    acc = np.full((len(tt), len(ff)), np.nan)

    # We will choose a "wavelet" that is essentially a Gaussian-modulated sinusoid.

    def window(x):
        return np.exp(-x**2/2)

    # Here we go!

    for j, ν in enumerate(tqdm(ff)):

        # The width of Gaussian modulation is chosen to be proportional
        # to the period of the modulated sinusoid, thus producing wavelet structure

        dt = Γ * (1/ν)
        for i, t in enumerate(tt):
            
            # we need to re-compute this for every combination of ν and t,
            # unfortunately.

            w = window((x-t)/dt)
            m = np.isfinite(np.nan_to_num(e_y/w, nan=np.inf))

            # actually compute the power: we use astropy's extirpolation-based L-S functionality.
            # rather than modifying the data points, we perform a weighted sum by
            # upweighting the uncertainties.

            # in principle we can get a massive performance improvement by using the NFFT
            # rather than the Press & Rybicki 1989 construction. Let's leave that for
            # later.

            ls = LombScargle(x[m], y[m], dy=(e_y/w)[m])
            p = float(ls.power(ν, normalization='psd'))

            # To preserve Parseval normalisation, we scale the power by the integral of the envelope,
            # which we can evaluate analytically. This allows us to recover a (blurred-out) Lomb-Scargle
            # periodogram when we integrate this with respect to time.

            p /= (np.sqrt(2 * np.pi) * dt)

            # Finally, we write this to the accumulating array.

            acc[i, j] = p
                
    return acc